;;
;;    TOUCH-SET!
;;

(macro (touch-set! form)
  `(_touch-set!_ ,(cadr form) ',(caddr form)))

(define set-promise-value! _touch-set!_)

;;
;; Parallel IF
;;

(macro (pif form)
  (let ((csq (gensym))
        (alt (gensym))
        (e1 (caddr form))
        (e2 (if (null? (cdddr form)) '() (cadddr form))))
    `(let ((,csq (future ,e1))
           (,alt (future ,e2)))
       (if ,(cadr form)
           (begin (touch-set! ,alt unspecified) ,csq)
           (begin (touch-set! ,csq unspecified) ,alt)))))

;;
;; Parallel OR
;;

(macro (exlambda form)
  (let ((l (gensym)))
    `(let ((,l (make-mutex)))
       (lambda ,(cadr form)
         (let ((r '()))
           (mutex-lock ,l)
           (set! r (begin ,@(cddr form)))
           (mutex-unlock ,l)
           r)))))

#|
(macro (por-old form)
  `(call/cc
     (lambda (return)
       (let ((el '())
             (len (length ',(cdr form))))
         (define lock (plambda (exp) (exp)))
         (define term
           (plambda ()
             (if (= len 1) (return #f))
             (set! len (1- len))))
         (define (kill p l)
           (cond ((null? l) #t)
             ((eq? (cdr l) p) (kill p (cdr l)))
             (else (touch-set! (car l) unspecified)
                   (kill p (cdr l)))))
         (define (f x)
           (lock
             (lambda ()
               (set! el 
                 (cons (let ((p el))
                         (future
                           (let ((ret (x)))
                             (if ret
                                 (lock
                                   (lambda () (kill p el)
                                              (return ret)))
                                 (term)))))
                       el)))))
         ,@(map (lambda (x) `(f (lambda () ,x)))
                (cdr form))
;        (suspend)))))
         (quit)))))
|#
#|
(macro (por form)
  `(let ((el '())
         (ans (make-placeholder))
         (len (length ',(cdr form))))
     (define lock (exlambda (exp) (exp)))
     (define term
       (exlambda ()
         (if (= len 1)
             (touch-set! ans #f)
             (set! len (1- len)))))
     (define (kill p l)
       (cond ((null? l) #t)
             ((eq? (cdr l) p) (kill p (cdr l)))
             (else (touch-set! (car l) unspecified)
                   (kill p (cdr l)))))
     (define (f x)
       (set! el
         (cons (let ((p el))
                 (future
                   (let ((ret (x)))
                     (if ret
                         (lock
                           (lambda () (kill p el)
                                      (set-promise-value! ans ret)
                                      (quit)))
                         (term)))))
               el)))
     (if (null? ',(cdr form))
         #f
         (begin
           (lock (lambda ()
                   ,@(map (lambda (x) `(f (lambda () ,x)))
                          (cdr form))))
           (touch ans)))))
|#
;;
;; Parallel AND
;;

#|
(macro (pand-old form)
  `(call/cc
     (lambda (return)
       (let ((el '())
             (len (length ',(cdr form))))
         (define lock (plambda (exp) (exp)))
         (define term
           (plambda (ret)
             (if (= len 1) (return ret))
             (set! len (1- len))))
         (define (kill p l)
           (cond ((null? l) #t)
                 ((eq? (cdr l) p) (kill p (cdr l)))
                 (else (touch-set! (car l) unspecified)
                       (kill p (cdr l)))))
         (define (f x)
           (lock
             (lambda ()
               (set! el 
                 (cons (let ((p el))
                         (future
                           (let ((ret (x)))
                             (if (not ret)
                                 (lock
                                   (lambda () (kill p el) (return #f)))
                                 (term ret)))))
                       el)))))
         ,@(map (lambda (x) `(f (lambda () ,x)))
                (cdr form))
;        (suspend)))))
         (quit)))))
|#
#|
(macro (pand form)
  `(let ((el '())
         (ans (make-placeholder))
         (len (length ',(cdr form))))
     (define lock (exlambda (exp) (exp)))
     (define term
       (exlambda (ret)
         (if (= len 1)
             (set-promise-value! ans ret)
             (set! len (1- len)))))
     (define (kill p l)
       (cond ((null? l) #t)
             ((eq? (cdr l) p) (kill p (cdr l)))
             (else (touch-set! (car l) unspecified)
                   (kill p (cdr l)))))
     (define (f x)
       (set! el 
         (cons (let ((p el))
                 (future
                   (let ((ret (x)))
                     (if (not ret)
                         (lock
                           (lambda () (kill p el)
                                      (touch-set! ans #f)
                                      (quit)))
                         (term ret)))))
               el)))
     (if (null? ',(cdr form))
         #t
         (begin
           (lock (lambda ()
                   ,@(map (lambda (x) `(f (lambda () ,x)))
                          (cdr form))))
           (touch ans)))))
|#
;;
;;    PLAMBDA for process closure
;;

(define (make-empty-queue)
  (let ((tmp (cons '() '())))
    (set-car! tmp tmp)
    (cons (make-mutex) tmp)))

(define (empty-queue? queue) (eq? (car queue) queue))

(define (enqueue queue x)
  (let ((tmp (cons x '())))
    (mutex-lock (car queue))
    (let ((q (cdr queue)))
      (if (empty-queue? q)
          (begin
            (set-cdr! (car q) tmp)
            (set-car! q tmp)
            (condition-signal (car queue) #t))
          (begin
            (set-cdr! (car q) tmp)
            (set-car! q tmp))))
    (mutex-unlock (car queue))))

(define (dequeue queue)
  (mutex-lock (car queue))
  (if (empty-queue? (cdr queue))
      (condition-wait (car queue)))
  (let ((q (cdr queue)))
    (let ((tmp (cdr q)))
      (if (eq? (car q) tmp)
          (begin
            (set-car! q q)
            (set-cdr! q '()))
          (set-cdr! q (cddr q)))
      (mutex-unlock (car queue))
      (car tmp))))

(macro (plambda form)
  `(let ((argument-queue (make-empty-queue))
         (the-task '()))
     (define (thunk)
       (let ((arglist (dequeue argument-queue)))
         (set-promise-value! (car arglist)
                             (apply (lambda ,@(cdr form))
                                    (cdr arglist))))
       (thunk))
     (set! the-task (future (thunk)))
     (lambda ,(cadr form)
       (let ((answer (make-placeholder)))
         (enqueue argument-queue
                  (list answer ,@(cadr form)))
         answer))))


;;
;;    PREMATURE-RETURN
;;

(define (p-return-aux plist alist)
  (if (null? plist)
      (list (car alist))
      (cons `(spawn-thread ,(car plist)
                           ,(car alist))
             (p-return-aux (cdr plist) (cdr alist)))))

(macro (premature-return form)
  (let* ((f (cadr form))
         (args (cddr form))
         (proms (map (lambda (x) '(make-placeholder)) args))
         (lamargs (map (lambda (x) (gensym)) (cdr args)))
         (lastarg (gensym)))
    `((lambda ,(cons lastarg lamargs)
        (xspawn-thread ,lastarg
                       ,(cons 'begin (p-return-aux `,lamargs `,args)))
        (,f ,@lamargs ,lastarg))
      ,@proms)))

;;
;;    CALL-WITH-PROMISE
;;

(define (call/p f)
  (let ((p (make-placeholder)))
    (f (lambda (v) (set-promise-value! p v)))
    (touch p)))

;; This is example of call/p.

#|
(define (foo)
  (future
    (let ((p (call/p bar)))
      (do () (#f)
        (format #t "~s " (thread-id))
        (flush-port (current-output-port))
        (do ((i 0 (1+ i))) ((> i 10000)))
        (set! p (call/p p))))))

(define (bar p)
  (future
    (do () (#f)
      (format #t "~s " (thread-id))      
      (flush-port (current-output-port))
      (do ((i 0 (1+ i))) ((> i 10000)))
      (set! p (call/p p)))))
|#

;;
;;    Load based FUTURE
;;

(macro (lfuture form)
  `(if (> (the-number-of-current-thread) 8)
       ,(cadr form)
       (future ,(cadr form))))

;;
;;    Parallel MAP
;;

(define (pmap proc lst)
  (if (pair? lst)
      (let ((tail (lfuture (pmap proc (cdr lst))))
            (val (proc (car lst))))
        (cons val tail))
;       (       ; for balance
;        (cons val (touch tail)))
      '()))

;;
;;    Parallel vector MAP
;;

(define (pvmap! proc vect)
  (define (map-range! lo hi)
    (if (= lo hi)
        (vector-set! vect lo (proc (vector-ref vect lo)))
        (let* ((mid (quotient (+ lo hi) 2))
               (sync (lfuture (map-range! (1+ mid) hi))))
          (map-range! lo mid)
          (touch sync))))
  (map-range! 0 (1- (vector-length vect)))
  vect)
