;;;
;;; REPL --- Readline version  by T.Komiya
;;;

(define *repl:columns* 80)

(define _read_ read)
(define console-input-port-eof? #f)

(let ((last-exp #f))
  (define (r . port)
    (let ((x (apply _read_ port)))
      (if (eq? (if (null? port) (current-input-port) (car port))
               (console-input-port))
          (set! last-exp x))
      x))
  (define (cipe?) (eof-object? last-exp))
  (set! read r)
  (set! console-input-port-eof? cipe?))

(define repl
  (let* ((byebye bye)
         (prompt "> ")
         (prompt-len (string-length prompt)))
    (define (pad-sp str)
      (make-string (- *repl:columns* 
                      (modulo (+ 2 (string-length str)) *repl:columns*))
                   #\space))
    (lambda ()
      (set! *trace-level* 0)
      (newline)
      (and (console-input-port-eof?) (byebye))
      (if (bound? 'tk:update) (tk:update)) ; for TUS/Tk
      (let ((line (readline prompt)))
        (if (eof-object? line)
            (begin (newline) (byebye)))
        (let ((expr (let loop ((l line))
                      (let ((x (repl:expr? l)))
                        (cond ((not x)
                               (loop
                                 (string-append
                                   l (pad-sp l) (readline ""))))
                              ((number? x)
                               (loop
                                 (let ((x (substring l 0 x)))
                                   (string-append x
                                     (pad-sp x) (readline "")))))
                              ((eq? x 'incomplete-string)
                               (loop
                                 (string-append
                                   l (string #\newline) (readline ""))))
                              (else l))))))
          (add-history expr)
          (write (eval (_read_ (open-input-string expr)))))
        (repl)))))

;;;
;;; (repl:expr? <string>)
;;;   - Return #t if <string> contains an expression.
;;;   - Return position of `;' if <string> contains comment.
;;;   - Return symbol `incomplete-string' if <string> contains incomplete
;;;     string
;;;
(define (repl:expr? str)
  (let ((i 0)
        (len (string-length str)))
    (define (getc)
      (and (< i len)
           (let ((c (string-ref str i)))
             (set! i (1+ i))
             c)))
    (define (ungetc) (set! i (1- i)))
    (define skip-ws
      (let ((flag #f))
        (lambda ()
          (if flag
              flag
              (let loop ((c (getc)))
                (cond ((not c) c)
                      ((char=? c #\;)
                       (let ((x (1- i)))
                         (set! i len)
                         (set! flag x)
                         x))
                      ((not (char-whitespace? c)) c)
                      (else (loop (getc)))))))))

    (define (readexp)
      (let ((c (skip-ws)))
        (cond ((not c) #f)
              ((number? c) 0)
              (else (ungetc) (readexp2)))))
    (define (readexp2)
      (let ((c (skip-ws)))
        (case c
          ((#\() (read-pair))  ; ) for balance
          ((#\)) #t)           ; ( for balance
          ((#\') (readexp2))
          ((#\`) (readexp2))
          ((#\,)
           (let ((c (getc)))
             (if (char=? c #\@)
                 (readexp2)
                 (begin (ungetc) (readexp2)))))
          ((#\#)
           (let ((c (getc)))
             (case c
               ((#\t #\f #\T #\F) #t)
               ((#\\) (getc) #t)  ;; NOTE: #\ is equivalent to #\space
               ((#\() (read-pair))  ; ) for balance
               ((#\b #\o #\d #\x #\i #\e #\B #\O #\D #\X #\I #\E)
                (readexp2))
               (else
                 (let loop ((c (getc)))
                   (if (or (not c)
                           (char=? c #\#)
                           (char=? c #\=))
                       #t
                       (loop (getc))))))))
          ((#\") (read-string))
          (else
            (if (number? c)
                c
                (or (not c)
                    (let loop ((c (getc)))
                      (cond ((not c) #t)
                            ((char-whitespace? c) #t)
                            ((or (char=? c #\()  ; ) for balance
                                 (char=? c #\))  ; ( for balance
                                 (char=? c #\')
                                 (char=? c #\`))
                             (ungetc)
                             #t)
                            ((char=? c #\;)
                             (ungetc)
                             #t)
                            (else (loop (getc)))))))))))
    (define (read-pair)
      (let ((c (skip-ws)))
        (cond ((not c) #f)
              ((number? c) c)
              ((char=? c #\)) #t) ; ( for balance
              (else
                (ungetc)
                (readexp2)
                (read-pair)))))
    (define (read-string)
      (let loop ((c (getc)))
        (cond ((not c) 'incomplete-string)
              ((char=? c #\") #t)
              (else
                (and (char=? c #\\) (getc))
                (loop (getc))))))
    (readexp)))

(define sctop-old sctop)
(set! sctop repl)
(sctop)
